/*
Class describing the bandwidth changes during the streaming
*/
#include "StdAfx.h"
#include "BwChanges.h"

/*
creates an empty object
*/
BwChanges::BwChanges(void)
{
	length = 0;
	currentTime = -1;
	currentSegment = -1;
}

/*
Creates on object with constant bandwidth
b: the bandwidth
*/
BwChanges::BwChanges(int b)
{
	length = 1;
	timeslots[0] = 0;
	bws[0] = b;
	currentTime = 0;
	currentSegment = 0;
};

/*
Creates on object with changes in the bandwidth
l: number of changes
times: the time slots of the changes
b: the bandwidths after the changes
*/
BwChanges::BwChanges(int l, int times[], int bws[])
{
	length = l;
	for(int i = 0; i < length; i++) {
		timeslots[i] = times[i];
		bws[i] = bws[i];
	}
	currentTime = 0;
	currentSegment = 0;
};

BwChanges::~BwChanges(void)
{
}

/*
returns with the bandwidth at a given time slot
*/
int BwChanges::getBw(int time) {
	int t = 0;

	while(t < length - 1 && timeslots[t+1] <= time) {
		t++;
	}
	return bws[t];
}

/*
sets the current time slot to the beginning of the video
*/
void BwChanges::resetTime() {
	currentTime = 0;
	currentSegment = 0;
}

/*
returns with the inital bandwidth
*/
int BwChanges::getFirst() {
	if(length == 0) {
		return -1;
	}
	currentTime = 0;
	currentSegment = 0;
	return bws[currentSegment];
}

/*
increases the current time slot by one and returns with the bandwidth at that time slot
*/
int BwChanges::getNext() {
	if(length == 0) {
		return -1;
	}
	if(currentSegment < length-1 && timeslots[currentSegment + 1] <= ++currentTime) {
		currentSegment++;
	}
	return bws[currentSegment];
}
