/* 
PiecePicker.cpp : Defines the entry point for the console application. It contains reading the input, 
calling the piece selection algorithms and writing the statsitics on the screen.
*/
#include "stdafx.h"

#include <iostream>
#include <random>

#include "common.h"
#include "SVCVideoStatus.h"
#include "BwChanges.h"

#define piece2quality(p, windowSize) (p)/(windowSize)
#define piece2ts(p, windowSize) (p)%(windowSize)

const char inputName[] = "piecepicking.txt";	// the file containing the input data
const int bufLength = 128;						// the length of a line read from the input
const int upSwitch = 5;			// not used - the length of the period with enough bandwidth capacity to switch up after that the switch up is allowed 

/*
reads input from parameters from a file
*/
int	readsInput(int lbw[], int &nTimeSlots, int &nQualities, int &windowSize, double &p, int &setupLength, BwChanges *bws) {
	FILE *fin;
	char buf[bufLength];
	float f;

	fopen_s(&fin, inputName, "r");
	if (fin == NULL)
		return -1;
	fgets(buf, bufLength, fin);
	sscanf_s(buf, "%d%s", &nQualities);
	fgets(buf, bufLength, fin);
	sscanf_s(buf, "%d%s", &nTimeSlots);
	fgets(buf, bufLength, fin);
	sscanf_s(buf, "%d%s", &windowSize);
	fgets(buf, bufLength, fin);
	sscanf_s(buf, "%d%s", &setupLength);
	for(int i = 0; i < nQualities; i++) {
		fscanf_s(fin, "%d", lbw + i);
	}
	fgets(buf, bufLength, fin); // reads the remarks
	fgets(buf, bufLength, fin);
	sscanf_s(buf, "%d%s", &bws->length);
	for(int i = 0; i < bws->length; i++) {
		fscanf_s(fin, "%d", bws->timeslots + i);
	}
	fgets(buf, bufLength, fin); // reads the remarks
	for(int i = 0; i < bws->length; i++) {
		fscanf_s(fin, "%d", bws->bws + i);
	}
	fgets(buf, bufLength, fin); // reads the remarks
	fgets(buf, bufLength, fin);
	sscanf_s(buf, "%f%s", &f);
	p = (double)f;

	fclose(fin);
	return 0;
}

/*
initalizations
*/
void init(int nQualities, int nTimeSlots, int lbw[maxNQualities], int c[maxNQualities][maxNTimeSlots], bool downloaded[maxNQualities][maxNTimeSlots], int &lastQuality, int &maxQuality, int switchCounter[])
{
    for( int i = 0; i < nQualities; i++) {
        for(int j = 0; j < nTimeSlots; j++) {
			c[i][j] = lbw[i];
			downloaded[i][j] = false;
		}
		switchCounter[i] = 0;
	}
	lastQuality = nQualities;	//it allows switch to any layer
	maxQuality = nQualities;
}

/*
Calculates delays of the pieces
p: the probability of finishing the download in one time slot 
*/
void initDelays(int nQualities, int nTimeSlots, double p, int delays[maxNQualities][maxNTimeSlots])
{
	if(p > 0.0 && p < 1.0) {
		std::default_random_engine generator;
		std::geometric_distribution<int> distribution(p);

		for( int i = 0; i < nQualities; i++) {
			for(int j = 0; j < nTimeSlots; j++) {
				delays[i][j] = distribution(generator) + 1;
			}
		}
	} else if(p == 1.0) {
		for( int i = 0; i < nQualities; i++) {
			for(int j = 0; j < nTimeSlots; j++) {
				delays[i][j] = 1;
			}
		}
	} 
}

/* 
dynamic programming solution
int index = index of the item you need to decide to take or not (we start with the last element of the array and we work toward the first)
int size = size still available at the backpack
int weights[] = array with the weights of all items
int values[] = array with the values of all items
*/

double matrix[maxNPieces][maxSize+1];
bool picks[maxNPieces][maxSize+1];

void knapsack(int nQualities, int windowSize, int cts, int nTimeSlots, int size, int weights[maxNQualities][maxNTimeSlots], double values[maxNQualities][maxWindowSize], bool downloaded[maxNQualities][maxNTimeSlots], int p[]) {
    int max_ts, l, ts, ts0, item, index;
	double	newValue;

	max_ts = min(nTimeSlots - cts, windowSize);
	for( int i = 0; i < nQualities * max_ts; i++) {
		l = piece2quality(i, max_ts);
		ts = piece2ts(i, max_ts);
		ts0 = ts + cts;
		for (int k=0;k<=size;k++){
			if (weights[l][ts0]<=k){
				if(i == 0) {
					matrix[i][k] = values[l][ts];
					picks[i][k]=true;
				} else {
					newValue = values[l][ts] + matrix[i-1][k-weights[l][ts0]];
					if (newValue > matrix[i-1][k]) {
						picks[i][k]=true;
						matrix[i][k] = newValue;
					}
					else {
						picks[i][k]=false;
						matrix[i][k] = matrix[i-1][k];
					}
				}
			}
			else {
				picks[i][k] = false;
				if(i == 0) {
					matrix[i][k] = 0;
				} else {
					matrix[i][k] = matrix[i-1][k];
				}
			}
		}
    }
	index = 0;
    for(item = nQualities * max_ts - 1; item >= 0; item--) {
        if (picks[item][size]){
			l = piece2quality(item, max_ts);
			ts = piece2ts(item, max_ts) + cts;	
			if(!downloaded[l][ts]) {
				p[index++] = l*nTimeSlots + ts;
				size -= weights[l][ts];
			}
        }
    }
	p[index] = -1;		//the end of the array
}

/*
Returns with the bitrate of piece
*/
int getSize(int size, int delay) {
	if(delay == 0)
		return -1;	//error, delay is smaller than zero
	return (int)ceil((double)size / delay);
}

/*
runs offline greedy verison 
stalling possible
not neccsarilly optimal!!!
*/
void optOffline(int nQualities, int nTimeSlots, int setupLength, BwChanges *bws, int weights[maxNQualities][maxNTimeSlots], int delays[maxNQualities][maxNTimeSlots], int latestDownload[maxOfflineBuffer][maxSize], int latestDownloadIndices[maxOfflineBuffer])
{
	int piece, latest;
	int latestDownloadSizes[maxOfflineBuffer];
	int pieceSize, delay;
	bool isDownloadSuccessful[maxOfflineBuffer];

	for(int ts = 0; ts < maxOfflineBuffer; ts++) {
		latestDownloadIndices[ts] = 0;
		latestDownloadSizes[ts] = 0;
		isDownloadSuccessful[ts] = true;
	}
	for(int l = 0; l < nQualities; l++) {
		for(int ts = 0; ts < nTimeSlots; ts++) {
			if(isDownloadSuccessful[ts]) {
				piece = l * nTimeSlots + ts;
				delay = delays[l][ts];	
				pieceSize = getSize(weights[l][ts], delay);
				latest = ts - delay + 1 + setupLength;
				while(delay > 0 && latest >= 0) {	// looking for the time slot of downloading
					while(latestDownloadSizes[latest] + pieceSize > bws->getBw(latest) && latest >= 0) {
						latest--;
					}
					if(latest >= 0) {
						delay--;
					}
				}
				if(latest >= 0) {
					for(int i = latest; i < latest + delays[l][ts]; i++) {
						latestDownload[i][latestDownloadIndices[i]++] = piece;
						latestDownloadSizes[i] += pieceSize;
					}
				} else {
					isDownloadSuccessful[ts] = false;
				}
			}
		}
	}
}

/* 
dynamic programming solution for the offline problem
int size[] = avalaible bandwidth
int weights[] = array with the weights of all items
int values[] = array with the values of all items
*/

double matrixOffline[maxNTimeSlots][maxTotalSize+1];
int picksOffline[maxNTimeSlots][maxTotalSize+1];

void knapsackOffline(SVCVideoStatus vs, int nQualities, int nTimeSlots, BwChanges *bws, int weights[maxNQualities][maxNTimeSlots], int latestDownload[maxOfflineBuffer][maxSize], int latestDownloadIndices[maxOfflineBuffer]) {
    int layer, sumCapacity, sumSize, b, latest, pieceSize;
	double	newValue, sumValue;
	int latestDownloadSizes[maxOfflineBuffer];

	sumCapacity = 0;
	b = bws->getFirst();
	for(int ts = 0; ts < nTimeSlots; ts++) {
		sumCapacity += b;
		for (int k=0;k<=sumCapacity;k++){
			sumSize = 0;
			sumValue = 0.0;
			picksOffline[ts][k]=0;
			matrixOffline[ts][k] = (ts == 0) ? 0 : matrixOffline[ts-1][k];
			for(int l = 0; l <= nQualities; l++) {
				if (sumSize<=k){
					if(ts == 0) {
						matrixOffline[ts][k] = sumValue;
						picksOffline[ts][k]=l;
					} else {
						newValue = sumValue + matrixOffline[ts-1][k-sumSize];
						if (newValue > matrixOffline[ts][k]) {
							picksOffline[ts][k]=l;
							matrixOffline[ts][k] = newValue;
						}
					}
				}
				sumSize += weights[l][ts];
				sumValue += vs.layer_weighting[l];
			}
		}
		for (int k=sumCapacity+1; k <= maxTotalSize; k++){
			matrixOffline[ts][k] = matrixOffline[ts][sumCapacity];
			picksOffline[ts][k] = picksOffline[ts][sumCapacity];
		}
		b = bws->getNext();
    }
    for(int ts = 0; ts < nTimeSlots; ts++) {
		latestDownloadIndices[ts] = 0;
		latestDownloadSizes[ts] = 0;
	}
    for(int ts = nTimeSlots - 1; ts >= 0; ts--) {
		layer = picksOffline[ts][sumCapacity];
		for(int l = 0; l < layer; l++) {
			latest = ts;
			pieceSize = weights[l][ts];
			while(latestDownloadSizes[latest] + pieceSize > bws->getBw(latest) && latest >= 0) {
				latest--;						// looking for the time slot of downloading
			}
			if(latest >= 0) {
				latestDownload[latest][latestDownloadIndices[latest]++] = l*nTimeSlots + ts;
				latestDownloadSizes[latest] += pieceSize;
				sumCapacity -= pieceSize;
			} else {
				printf("Error in the offline algorithm!!!\n");
			}
	    }
    }
}

/*
makes a conversion between the different piece indexing methods
*/
int rescalePieceIndex(int piece, int cts, int l1, int l2) {
	return piece / l1 * l2 + piece % l1 - cts;
}

/*
runs piece-picking algorithm
*/
int *solve(SVCVideoStatus vs, int nQualities, int windowSize, int cts, int iSetup, int nTimeSlots, int setupLength, int b, int c[maxNQualities][maxNTimeSlots], bool downloaded[maxNQualities][maxNTimeSlots], int latestDownload[maxOfflineBuffer][maxSize], int latestDownloadIndices[maxOfflineBuffer], char cAlgorithm, tPieceUtility pieceUtility[], int p[]) 
{
	int piece, index, l, ts, max_ts, cts2;
	double utilities[maxNQualities][maxWindowSize];

	index = 0;
	switch(cAlgorithm) {
	case 'b':
		max_ts = min(nTimeSlots - cts, windowSize);
		for(l = 0; l < nQualities; l++) {
	        for(ts = cts; ts < cts + max_ts; ts++) {
		        piece = l * nTimeSlots + ts;
				if( !downloaded[l][ts] && b >= c[l][ts]) {
					p[index++] = piece;
					b -= c[l][ts];
				}
			}
		}
		p[index] = -1;		//the end of the array
		break;
	case 'g':
		vs.get_high_range(windowSize, cts, utilities);
		vs.get_high_range_kp2(windowSize, cts, c, true, utilities, pieceUtility);
		for(int i = 0; i < vs.lPieces; i++) {
			piece = pieceUtility[i].first;
			l = piece2quality(piece, nTimeSlots);
			ts = piece2ts(piece, nTimeSlots);
			if( !downloaded[l][ts] && b >= c[l][ts]) {
				p[index++] = piece;
				b -= c[l][ts];
			}
		}
		p[index] = -1;		//the end of the array
		break;
	case 'd':
		vs.get_high_range(windowSize, cts, utilities);
		knapsack(nQualities, windowSize, cts, nTimeSlots, b, c, utilities, downloaded, p);
		break;
	case 'o':
	case 'k':
		cts2 = (iSetup < 0) ? cts + setupLength : iSetup;
		for(int i = 0; i < latestDownloadIndices[cts2]; i++) {
			p[index++] = latestDownload[cts2][i];
		}
		p[index] = -1;		//the end of the array
		break;
	default:
		printf("Invalid code\n");
		break;
	}
	return p;
}

/*
updates the state of the pieces
*/
void update(int *p, int index, int nQualities, int nTimeSlots, int lbw[maxNQualities], int c[maxNQualities][maxNTimeSlots], int delays[maxNQualities][maxNTimeSlots], bool downloaded[maxNQualities][maxNTimeSlots])
{
	int i = 0, l, ts;

	while(p[i] != -1) {
		l = piece2quality(p[i], nTimeSlots);
		ts = piece2ts(p[i], nTimeSlots);
		c[l][ts] -= getSize(lbw[l], delays[l][ts]);			// the remaining size
		if(c[l][ts] <= 0) {
			c[l][ts] = 0;
			downloaded[l][ts] = true;
		} 
		i++;
	}
}

/*
Checks bw and allows switch up 
*/
bool bwMonitoring(char cAlgorithm, int &lastQuality, int nQualities, int currentLayer, int &maxQuality, int switchCounter[], bool &repeat)
{
	repeat = false;
	if(lastQuality < nQualities - 1 && !repeat && (cAlgorithm == 'g' || cAlgorithm == 'd')) {
		maxQuality = lastQuality;
		for(int j = lastQuality; j < nQualities; j++) {
			if(j < currentLayer) {
				switchCounter[j]++;
				if(switchCounter[j] >= upSwitch) {
					maxQuality = j;
				} 
			} else {
				switchCounter[j] = 0;
			}
		}
		repeat = (maxQuality != currentLayer);
	} 
	lastQuality = currentLayer;

	return repeat; 
}

/*
prints the state of the state of the pieces
*/
void putSolution(int nQualities, int windowSize, int cts, int nTimeSlots, int c[maxNQualities][maxNTimeSlots], int lbw[maxNQualities])
{
	int max_ts;

	printf("%d\n", cts);
	max_ts = (cts + windowSize > nTimeSlots) ? nTimeSlots : cts + windowSize;
	for( int i = nQualities - 1; i >= 0; i--) {
		for(int j = cts; j < max_ts; j++) {
			printf("%4d ", lbw[i] - c[i][j]);
		}
		std::cout<< '\n';
	}
}

/*
writer some statistics
*/
void createStatistics(int nQualities, int cts, bool downloaded[maxNQualities][maxNTimeSlots], int highestLayer[maxNTimeSlots])
{
	int highest = -1;

	for(int l = 0; l < nQualities; l++) {
		if(downloaded[l][cts]) {
			highest = l;
		} else {
			break;
		}
	}
	highestLayer[cts] = highest;
}

/*
writer some statistics
*/
void writeStatistics(SVCVideoStatus vs, char cAlgorithm, int nTimeSlots, int nQualities, int highestLayer[maxNTimeSlots])
{
	double utility = 0;
	double *utilities = new double[nQualities];

	utilities[0] = 0.0;
	for(int i = 0; i < nQualities; i++)  {
		utilities[i+1] = utilities[i] + vs.layer_weighting[i];
	}
	printf("statistics - algorithm: %c\n", cAlgorithm);
	for( int i = 0; i < nTimeSlots; i++) {
		printf("%3d ", i);
	}
		std::cout<< '\n';
	for( int i = 0; i < nTimeSlots; i++) {
		printf("%3d ", highestLayer[i]+1);
		utility += utilities[highestLayer[i]+1];
	}
	std::cout<< '\n';
	std::cout<< "Total utility:"<< utility <<'\n';
}

/*
The main program
*/
int _tmain(int argc, _TCHAR* argv[])
{
	int lbw[maxNQualities] = {100,100,100,100};	// the bw of a piece on a specific layer
	int nTimeSlots = 100;						// the length of the video in timeslots
	int nQualities = 4;							// the number of quality layers
	int windowSize = 20;						// the sliding window size
	int setupLength = 0;						// the length of the setup phase
	BwChanges *bws = new BwChanges(1000);	//the total bw at different times
	int b;
	double pr = 0.5;			//parameter of the geometric distribution
							
	char cAlgorithm;
	bool use_download_probability = true;

	int c[maxNQualities][maxNTimeSlots];	//the size of a piece
	bool downloaded[maxNQualities][maxNTimeSlots];	//true: fully downloaded, false not
	int delays[maxNQualities][maxNTimeSlots];	//realized delays of the pieces

	tPieceUtility *pieceUtility = new tPieceUtility[nQualities*windowSize];	//piece utility array
	int *p = new int[maxNPieces + 1];	//the selected pieces (with indexes)
	int latestDownload[maxOfflineBuffer][maxSize]; // for the optimal offline method
	int latestDownloadIndices[maxOfflineBuffer]; // the length of the latestDownload columns

	int highestLayer[maxNTimeSlots]; // quality produced by the algorithm
	int lastQuality;					// last quality for smoothing purposes
	int maxQuality;						// for smoothing purposes max quality can be introduced
	int switchCounter[maxNQualities];	// preservers the number of quality jumps

	int error;

	bool repeat;

	// reads input
	error = readsInput(lbw, nTimeSlots, nQualities, windowSize, pr, setupLength, bws);
	if(error < 0) {
		std::cout<< "Error in the input file! Default values are used.\n";
	}
	do {
		// Selects algorithm
		std::cout<< "Algorithm: ";
		cAlgorithm = std::getchar();
		if(cAlgorithm == 'x') {
			break;
		}
		if(cAlgorithm != 'b' && cAlgorithm != 'd' && cAlgorithm != 'g' && cAlgorithm != 'o' && cAlgorithm != 'k') {
			continue;
		}
		// Initialization
		SVCVideoStatus vs(nTimeSlots, nQualities, pr);
		init(nQualities, nTimeSlots, lbw, c, downloaded, lastQuality, maxQuality, switchCounter);
		initDelays(nQualities, nTimeSlots, pr, delays);
		if(cAlgorithm == 'o') {
			optOffline(nQualities, nTimeSlots, setupLength, bws, c, delays, latestDownload, latestDownloadIndices);
		} else if(cAlgorithm == 'k') {
			knapsackOffline(vs, nQualities, nTimeSlots, bws, c, latestDownload, latestDownloadIndices);
		}
		// Setup phase
		b = bws->getFirst();
		for(int i = 0; i < setupLength; i++) {	
			p = solve(vs, nQualities, windowSize, 0, i, nTimeSlots, setupLength, b, c, downloaded, latestDownload, latestDownloadIndices, cAlgorithm, pieceUtility, p);
			update(p, 0, nQualities, windowSize, lbw, c, delays, downloaded);
			putSolution(nQualities, windowSize, 0, nTimeSlots, c, lbw);
		}
		// Main part
		for(int i = 0; i < nTimeSlots; i++) {
			repeat = false;
			maxQuality = nQualities;
//			do {								// bandwidth monitoring is not used
				p = solve(vs, maxQuality, windowSize, i, -1, nTimeSlots, setupLength, b, c, downloaded, latestDownload, latestDownloadIndices, cAlgorithm, pieceUtility, p);
				update(p, i, nQualities, nTimeSlots, lbw, c, delays, downloaded);
				createStatistics(nQualities, i, downloaded, highestLayer);
//			} while(bwMonitoring(cAlgorithm, lastQuality, nQualities, highestLayer[i], maxQuality, switchCounter, repeat)); 
			putSolution(nQualities, windowSize, i, nTimeSlots, c, lbw);
			b = bws->getNext();
		}
		writeStatistics(vs, cAlgorithm, nTimeSlots, nQualities, highestLayer);
	} while(cAlgorithm != 'x');
}
