/*
Class describing a video
*/
#include "StdAfx.h"

#include <algorithm>

#include "SVCVideoStatus.h"


SVCVideoStatus::SVCVideoStatus(void)
{
}

/*
Creates a video status object
nTimeslots: the length of the video in timeslots
windowSize: the sliding window size
nQualities: the number of quality layers
pr: the probability of finishing the download in one time slot 
*/
SVCVideoStatus::SVCVideoStatus(int nTimeSlots, int nQualities, double pr)
{
    this->movie_numpieces = nTimeSlots;

	//the available layers in the torrent
    this->available_qualities = nQualities;

	this->avg_number_of_ts_per_piece = 1;

    //the layer weighting for the KP algorithm
	this->layer_weighting[0] = 8;
	this->layer_weighting[1] = 3;
	this->layer_weighting[2] = 2;
	this->layer_weighting[3] = 1;
    this->urgency_weighting = 2.0;
	this->use_download_probability = (pr != 1.0);       
}

SVCVideoStatus::~SVCVideoStatus(void)
{
}

/*
inits download probabilities
*/
void SVCVideoStatus::init_dp_array(int maxj)
{
	for(int i = 0; i < available_qualities; i++){
		for(int j = 0; j < maxj; j++){
			dp_array[i][j] = 1.0;
		}
	}
}

/*
calculates download probabilities
i: layer
j: timeslot
Precondition: the download probability on lower layers are already calculated
*/
double SVCVideoStatus::calculate_download_probability(int i, int j) {
		double bcr = (j + 1) / this->avg_number_of_ts_per_piece;
		double download_probability = 1.0;
		if( bcr > 1.5)
		    download_probability = 1.0;
		else if(bcr < 0.5)
		    download_probability = 0.0;
		else
		    download_probability = bcr - 0.5;
		if(i > 0) {
			download_probability = download_probability * this->dp_array[i-1][j];
		}
		this->dp_array[i][j] = download_probability;
		return download_probability;
}

/*
calculates utility
*/
void SVCVideoStatus::get_high_range(int windowSize, int cts, double utilities[maxNQualities][maxWindowSize])
{
	int index = 0, max_ts;
    double download_probability, priority, urgency;
	
	init_dp_array(windowSize);
	max_ts = min(movie_numpieces - cts, windowSize);
	for( int i = 0; i < available_qualities; i++) {
        for(int j = 0; j < max_ts; j++) {
            download_probability = 1.0;
			if(this->use_download_probability) {
				download_probability = calculate_download_probability(i, j);
			}
            priority = this->layer_weighting[i] * download_probability;
            urgency = pow((j + 1), this->urgency_weighting);
            utilities[i][j] = priority / urgency;
		}
	}
}

/*
compares two piece utility structure for sorting
*/
static bool functionSortPieceUtility (tPieceUtility x, tPieceUtility y) 
{ return (x.second>y.second); }

/*
Divedes the utilities with the sizes and sorts them for the greedy method
Returns [(first, last), (first, last), ..] ordered list of tuples
*/
void SVCVideoStatus::get_high_range_kp2(int windowSize, int cts, int c[maxNQualities][maxNTimeSlots], bool weighted, double utilities[maxNQualities][maxWindowSize], tPieceUtility utility_list[])
{
	int index = 0, piece_pos, max_ts;
    double utility_val;

	max_ts = min(movie_numpieces - cts, windowSize);
	for( int i = 0; i < available_qualities; i++) {
        for(int j = 0; j < max_ts; j++) {
            piece_pos = i * movie_numpieces + j +cts;
            utility_val = utilities[i][j];
			if(weighted && c[i][j + cts] > 0) {
				utility_val /= c[i][j + cts];
			}
			utility_list[index].first = piece_pos;
			utility_list[index++].second = utility_val;
		}
	}
    std::sort(utility_list, utility_list + index, functionSortPieceUtility);
	this->lPieces = index;
}
